"""Google provider code."""
import time
import googleapiclient.discovery
import yaml

from beeflow.common.cloud import provider


class GoogleProvider(provider.Provider):
    """Google provider class."""

    def __init__(self, project, zone, **kwargs):
        """Google provider constructor."""
        self.params = kwargs
        self.zone = zone
        self.project = project
        # Set defaults here for now
        self._api = googleapiclient.discovery.build('compute', 'v1')

    def get_ext_ip_addr(self, node_name):
        """Get the external IP of this node (or None if no IP)."""
        res = self._api.instances().get(instance=node_name,
                                        project=self.project,
                                        zone=self.zone).execute()
        try:
            return res['networkInterfaces'][0]['accessConfigs'][0]['natIP']
        except (IndexError, KeyError):
            return None

    def setup_cloud(self, config):
        """Set up the cloud based on the config information."""
        # Load the YAML data
        config = yaml.load(config, Loader=yaml.Loader)
        # This just creates instances one-by-one. There may be a better API call
        # to just create everything at once.
        for instance in config['instances']:
            call = self._api.instances().insert(project=self.project,
                                                zone=self.zone, body=instance)
            res = call.execute()
            print(res)
            time.sleep(2)
# Ignore E1101: Pylama notes that self._api doesn't have an 'instances' member.
#               I believe this is set at runtime by googleapiclient.
# pylama:ignore=E1101



"""Special connection for communicating with cli.py."""
from contextlib import contextmanager
import socket
import os
import traceback

import jsonpickle


class BeeflowConnectionError(Exception):
    """Connection error class."""

    def __init__(self):
        """Create a new connection error class."""


class Server:
    """Socket server for use in cli.py."""

    def __init__(self, s):
        """Create a new server."""
        self.s = s

    def accept(self):
        """Accept a new connection or return None."""
        try:
            conn, _ = self.s.accept()
            return ClientConnection(conn)
        except BlockingIOError:
            # No clients right now
            return None


class ClientConnection:
    """Connection to a client from the server end."""

    def __init__(self, conn):
        """Create a new connection."""
        self.conn = conn
        self.conn.setblocking(False)

    def get(self):
        """Get and return the message."""
        try:
            return _recv_message(self.conn)
        except (ConnectionResetError, OSError) as err:
            traceback.print_exc()
            print(err)
            raise BeeflowConnectionError() from err

    def put(self, msg):
        """Put a response message (closes the socket when done)."""
        try:
            _send_message(self.conn, msg)
            self.conn.close()
        except (ConnectionResetError, OSError) as err:
            traceback.print_exc()
            raise BeeflowConnectionError() from err


@contextmanager
def server(path):
    """Create a new server connection."""
    try:
        with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as s:
            s.bind(path)
            s.listen(1)
            s.setblocking(False)
            yield Server(s)
    finally:
        # Need to remove the socket when done
        os.remove(path)


def send(path, msg):
    """Send a single message to the server and get a response."""
    with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as s:
        try:
            s.connect(path)
        except (FileNotFoundError, ConnectionRefusedError):
            return None
        # Send everything
        _send_message(s, msg)
        return _recv_message(s)


SIZE_BYTE_COUNT = 4
SIZE_BYTE_ORDER = 'big'


def _send_message(s, msg):
    """Send a message on the socket."""
    data = bytes(jsonpickle.encode(msg), encoding='utf-8')
    size = len(data).to_bytes(SIZE_BYTE_COUNT, byteorder=SIZE_BYTE_ORDER)
    s.sendall(b''.join([size, data]))


def _recv_message(s):
    """Receive a message on the socket."""
    # Get the message size
    size = []
    while len(size) < SIZE_BYTE_COUNT:
        buf = s.recv(SIZE_BYTE_COUNT)
        size.extend(buf)
    size = bytes(size)
    size = int.from_bytes(size, byteorder=SIZE_BYTE_ORDER)
    # Now get the message data
    data = []
    while len(data) < size:
        buf = s.recv(size)
        data.extend(buf)
    data = bytes(data)
    return jsonpickle.decode(str(data, encoding='utf-8'))



#! /usr/bin/env python3
"""Unit test module for the BEE workflow interface module."""

# Disable protected member access warning
# pylama:ignore=W0212

import unittest

from beeflow.common.config_driver import BeeConfig as bc

bc.init()

from beeflow.common.wf_data import (Requirement, Hint, InputParameter, OutputParameter,
                                    StepInput, StepOutput, generate_workflow_id)
from beeflow.common.wf_interface import WorkflowInterface
from beeflow.tests.mocks import MockGDBInterface


class TestWorkflowInterface(unittest.TestCase):
    """Unit test case for the workflow interface."""

    @classmethod
    def setUpClass(cls):
        """Start the GDB and initialize the Workflow interface."""
        mock_gdb_iface = MockGDBInterface()
        cls.wfi = WorkflowInterface(mock_gdb_iface)

    def tearDown(self):
        """Clear all data in the Neo4j database."""
        if self.wfi.workflow_initialized() and self.wfi.workflow_loaded():
            self.wfi.finalize_workflow()

    def test_initialize_workflow(self):
        """Test workflow initialization.

        The workflow node and associated requirement/hint nodes should have been created.
        """
        requirements = [Requirement("ResourceRequirement", {"ramMin": 1024})]
        hints = [Hint("ResourceRequirement", {"ramMin": 1024}),
                 Hint("NetworkAccess", {"networkAccess": True})]
        workflow_id = generate_workflow_id()
        workflow = self.wfi.initialize_workflow(
            workflow_id, "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "viz/output")],
            requirements, hints)

        gdb_workflow, _ = self.wfi.get_workflow()

        self.assertEqual(workflow, gdb_workflow)
        self.assertEqual(workflow_id, workflow.id)
        self.assertEqual(workflow.id, gdb_workflow.id)
        self.assertIsNotNone(self.wfi._workflow_id)
        self.assertEqual("SUBMITTED", self.wfi.get_workflow_state())

    def test_execute_workflow(self):
        """Test workflow execution initialization (set initial tasks' states to 'READY')."""
        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "viz/output")])
        tasks = self._create_test_tasks()
        self.wfi.execute_workflow()

        self.assertEqual("READY", self.wfi.get_task_state(tasks[0]))
        self.assertEqual("RUNNING", self.wfi.get_workflow_state())

    def test_pause_workflow(self):
        """Test workflow execution pausing (set running tasks' states to 'PAUSED')."""
        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "viz/output")])
        self._create_test_tasks()

        self.wfi.execute_workflow()

        self.wfi.pause_workflow()

        # Workflow state should now be 'PAUSED'
        self.assertEqual("PAUSED", self.wfi.get_workflow_state())

    def test_resume_workflow(self):
        """Test workflow execution resuming (set paused tasks' states to 'RUNNING')."""
        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "viz/output")])
        self._create_test_tasks()

        self.wfi.execute_workflow()
        self.wfi.pause_workflow()
        self.wfi.resume_workflow()

        # Workflow state should now be 'RESUME'
        self.assertEqual("RESUME", self.wfi.get_workflow_state())

    def test_reset_workflow(self):
        """Test workflow execution resetting (set all tasks to 'WAITING', delete metadata)."""
        workflow = self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "viz/output")])
        tasks = self._create_test_tasks()
        metadata = {"cluster": "fog", "crt": "charliecloud",
                    "container_md5": "67df538c1b6893f4276d10b2af34ccfe", "job_id": 1337}

        # Set tasks' metadata, set state to COMPLETED
        for task in tasks:
            self.wfi.set_task_metadata(task, metadata)
            self.wfi.set_task_state(task, "COMPLETED")
            self.assertEqual("COMPLETED", self.wfi.get_task_state(task))

        workflow_id = 42
        self.wfi.reset_workflow(workflow_id)

        # States should be reset, metadata should be deleted
        for task in tasks:
            self.assertDictEqual({}, self.wfi.get_task_metadata(task))
            self.assertEqual("WAITING", self.wfi.get_task_state(task))

        # Workflow ID should be reset
        (gdb_workflow, gdb_tasks) = self.wfi.get_workflow()
        new_workflow_id = gdb_workflow.id

        self.assertNotEqual(new_workflow_id, workflow.id)
        for task in gdb_tasks:
            self.assertEqual(task.workflow_id, new_workflow_id)

    def test_finalize_workflow(self):
        """Test workflow deletion from the graph database."""
        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "viz/output")])
        self.wfi.finalize_workflow()

        self.assertFalse(self.wfi.workflow_loaded())

    def test_add_task(self):
        """Test task creation."""
        task_name = "test_task"
        base_command = ["ls", "-a", "-F"]
        inputs = [StepInput("test_input", "File", "input.txt", "default.txt", "test_input", "-l",
                            None, None)]
        outputs = [StepOutput("test_input/test_task_done", "stdout", "output.txt", "output.txt")]
        requirements = [Requirement("ResourceRequirement", {"ramMin": 1024}),
                        Requirement("NetworkAccess", {"networkAccess": True})]
        hints = [Hint("ResourceRequirement", {"ramMin": 1024}),
                 Hint("NetworkAccess", {"networkAccess": True})]
        stdout = "output.txt"
        stderr = "output-err.txt"

        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "test_input/test_task_done")])
        task = self.wfi.add_task(
            name=task_name,
            base_command=base_command,
            inputs=inputs,
            outputs=outputs,
            requirements=requirements,
            hints=hints,
            stdout=stdout,
            stderr=stderr)

        # Task object assertions
        self.assertEqual(task_name, task.name)
        self.assertEqual(base_command, task.base_command)
        self.assertCountEqual(inputs, task.inputs)
        self.assertCountEqual(outputs, task.outputs)
        self.assertCountEqual(requirements, task.requirements)
        self.assertCountEqual(hints, task.hints)
        self.assertEqual(stdout, task.stdout)
        self.assertEqual(stderr, task.stderr)
        self.assertEqual(task.workflow_id, self.wfi._workflow_id)
        self.assertIsInstance(task.id, str)

        # Graph database assertions
        gdb_task = self.wfi.get_task_by_id(task.id)
        self.assertEqual(task, gdb_task)
        self.assertEqual(task.id, gdb_task.id)

    def test_restart_task(self):
        """Test restart of failed task."""
        task_name = "test_task"
        base_command = ["ls", "-a", "-F"]
        inputs = [StepInput("test_input", "File", "input.txt", "default.txt", "test_input", "-l",
                            None, None)]
        outputs = [StepOutput("test_input/test_task_done", "stdout", "output.txt", "output.txt")]
        requirements = [Requirement("ResourceRequirement", {"ramMin": 1024}),
                        Requirement("NetworkAccess", {"networkAccess": True})]
        hints = [Hint("ResourceRequirement", {"ramMin": 1024}),
                 Hint("NetworkAccess", {"networkAccess": True}),
                 Hint("beeflow:CheckpointRequirement", {"file_path": "checkpoint_output",
                                                        "container_path": "checkpoint_output",
                                                        "file_regex": "backup[0-9]*.crx",
                                                        "restart_parameters": "-R",
                                                        "num_tries": 2})]
        stdout = "output.txt"
        stderr = "output-err.txt"
        test_checkpoint_file = "/backup0.crx"

        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "test_input/test_task_done")])
        task = self.wfi.add_task(
            name=task_name,
            base_command=base_command,
            inputs=inputs,
            outputs=outputs,
            requirements=requirements,
            hints=hints,
            stdout=stdout,
            stderr=stderr)

        # Restart the task, should create a new Task
        new_task = self.wfi.restart_task(task, test_checkpoint_file)

        # Assert inequality of Task objects
        self.assertNotEqual(task.id, new_task.id)
        self.assertEqual("test_task(1)", new_task.name)

        # Assert equality of graph database objects
        self.assertEqual(new_task, self.wfi.get_task_by_id(new_task.id))
        self.assertEqual("RESTARTED", self.wfi.get_task_state(task))
        self.assertEqual("READY", self.wfi.get_task_state(new_task))
        self.assertEqual(self.wfi.get_task_metadata(task),
                         self.wfi.get_task_metadata(new_task))

        # Check that task command includes checkpoint file
        print(new_task.command)
        self.assertListEqual(['ls', '-a', '-F', '-l', 'input.txt', '-R',
                              '/backup0.crx'], new_task.command)

        # Restart once again
        newer_task = self.wfi.restart_task(new_task, test_checkpoint_file)

        # Assert inequality of Task objects
        self.assertNotEqual(new_task.id, newer_task.id)
        self.assertEqual("test_task(2)", newer_task.name)

        # Assert equality of graph database objects
        self.assertEqual(newer_task, self.wfi.get_task_by_id(newer_task.id))
        self.assertEqual("RESTARTED", self.wfi.get_task_state(new_task))
        self.assertEqual("READY", self.wfi.get_task_state(newer_task))
        self.assertEqual(self.wfi.get_task_metadata(new_task),
                         self.wfi.get_task_metadata(newer_task))

        # Restart on more time (should return None)
        self.assertIsNone(self.wfi.restart_task(newer_task, test_checkpoint_file))

        # Check that task command includes checkpoint file
        self.assertListEqual(['ls', '-a', '-F', '-l', 'input.txt', '-R',
                              '/backup0.crx'], newer_task.command)

    def test_finalize_task(self):
        """Test finalization of completed tasks."""
        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "viz/output")])
        tasks = self._create_test_tasks()
        self.wfi.execute_workflow()
        self.wfi.set_task_output(tasks[0], "prep/prep_output", "prep_output.txt")
        ready_tasks = self.wfi.finalize_task(tasks[0])

        # Manually set expected task input values for comparison
        tasks[1].inputs = [StepInput("input_data", "File", "prep_output.txt", None,
                                     "prep/prep_output", None, None, None)]
        tasks[2].inputs = [StepInput("input_data", "File", "prep_output.txt", None,
                                     "prep/prep_output", None, None, None)]
        tasks[3].inputs = [StepInput("input_data", "File", "prep_output.txt", None,
                                     "prep/prep_output", None, None, None)]

        self.assertCountEqual(tasks[1:4], ready_tasks)

    def test_get_task_by_id(self):
        """Test obtaining a task from the graph database by its ID."""
        task_name = "test_task"
        base_command = "ls"
        inputs = [StepInput("input", "File", "input.txt", "default.txt", "test_input", None, None,
                            None)]
        outputs = [StepOutput("test_task/test_task_done", "stdout", "output.txt", "output.txt")]
        requirements = [Requirement("ResourceRequirement", {"ramMin": 1024}),
                        Requirement("NetworkAccess", {"networkAccess": True})]
        hints = [Hint("ResourceRequirement", {"ramMin": 1024}),
                 Hint("NetworkAccess", {"networkAccess": True})]
        stdout = "output.txt"
        stderr = "output-err.txt"

        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "test_task/test_task_done")])
        task = self.wfi.add_task(
            name=task_name,
            base_command=base_command,
            inputs=inputs,
            outputs=outputs,
            requirements=requirements,
            hints=hints,
            stdout=stdout,
            stderr=stderr)

        self.assertEqual(task, self.wfi.get_task_by_id(task.id))

    def test_get_workflow(self):
        """Test obtaining the workflow from the graph database."""
        requirements = [Requirement("ResourceRequirement", {"ramMin": 1024})]
        hints = [Hint("NetworkAccess", {"networkAccess": True})]
        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "viz/output")],
            requirements, hints)
        tasks = self._create_test_tasks()

        (workflow, wf_tasks) = self.wfi.get_workflow()

        self.assertCountEqual(tasks, wf_tasks)
        self.assertCountEqual(requirements, workflow.requirements)
        self.assertCountEqual(hints, workflow.hints)

    def test_get_workflow_outputs(self):
        """Test obtaining the outputs of a workflow."""
        workflow = self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "test_task/test_task_done")])

        self.assertListEqual(workflow.outputs, self.wfi.get_workflow_outputs())

    def test_get_workflow_state(self):
        """Test obtaining the state of a workflow."""
        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "test_task/test_task_done")])

        # Initialized workflow state should be 'SUBMITTED'
        self.assertEqual("SUBMITTED", self.wfi.get_workflow_state())

        self.wfi.execute_workflow()

        # Executed workflow state should be 'RUNNING'
        self.assertEqual("RUNNING", self.wfi.get_workflow_state())

    def test_set_workflow_state(self):
        """Test setting the state of a workflow."""
        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "test_task/test_task_done")])

        # Initialized workflow state should be 'SUBMITTED'
        self.assertEqual("SUBMITTED", self.wfi.get_workflow_state())

        self.wfi.set_workflow_state("RUNNING")

        # Workflow state should now be 'RUNNING'
        self.assertEqual("RUNNING", self.wfi.get_workflow_state())

    def test_get_ready_tasks(self):
        """Test obtaining of ready workflow tasks."""
        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "viz/output")])
        tasks = self._create_test_tasks()

        # Should be no ready tasks
        self.assertListEqual([], self.wfi.get_ready_tasks())

        # Set Compute tasks to READY
        for task in tasks[1:4]:
            self.wfi.set_task_state(task, "READY")

        self.assertCountEqual(tasks[1:4], self.wfi.get_ready_tasks())

    def test_get_dependent_tasks(self):
        """Test obtaining of dependent tasks."""
        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "viz/output")])
        tasks = self._create_test_tasks()
        # Get dependent tasks of Data Prep
        dependent_tasks = self.wfi.get_dependent_tasks(tasks[0])

        # Should equal Compute 0, Compute 1, and Compute 2
        self.assertCountEqual(tasks[1:4], dependent_tasks)

    def test_get_task_state(self):
        """Test obtaining of task state."""
        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "test_task/output")])
        task = self.wfi.add_task(
            "test_task",
            "ls",
            [StepInput("test_input", "File", "input.txt", "default.txt", "test_input", None, None,
                       None)],
            [StepOutput("test_task/output", "File", "output.txt", "output.txt")])

        # Should be WAITING because workflow not yet executed
        self.assertEqual("WAITING", self.wfi.get_task_state(task))

    def test_set_task_state(self):
        """Test the setting of task state."""
        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "test_task/output")])
        task = self.wfi.add_task(
            "test_task",
            "ls",
            [StepInput("test_input", "File", "input.txt", "default.txt", "test_input", None, None,
                       None)],
            [StepOutput("test_task/output", "File", "output.txt", "output.txt")])

        self.wfi.set_task_state(task, "RUNNING")

        # Should now be RUNNING
        self.assertEqual("RUNNING", self.wfi.get_task_state(task))

    def test_get_task_metadata(self):
        """Test the obtaining of task metadata."""
        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "test_task/output")])
        task = self.wfi.add_task(
            "test_task",
            "ls",
            [StepInput("test_input", "File", "input.txt", "default.txt", "test_input", None, None,
                       None)],
            [StepOutput("test_task/output", "File", "output.txt", "output.txt")])
        metadata = {"cluster": "fog", "crt": "charliecloud",
                    "container_md5": "67df538c1b6893f4276d10b2af34ccfe", "job_id": 1337}

        self.wfi.set_task_metadata(task, metadata)
        self.assertDictEqual(metadata, self.wfi.get_task_metadata(task))

    def test_set_task_metadata(self):
        """Test the setting of task metadata."""
        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "test_task/output")])
        task = self.wfi.add_task(
            "test_task",
            "ls",
            [StepInput("test_input", "File", "input.txt", "default.txt", "test_input", None, None,
                       None)],
            [StepOutput("test_task/output", "File", "output.txt", "output.txt")])
        metadata = {"cluster": "fog", "crt": "charliecloud",
                    "container_md5": "67df538c1b6893f4276d10b2af34ccfe", "job_id": 1337}

        # Metadata should be empty
        self.assertDictEqual({}, self.wfi.get_task_metadata(task))

        self.wfi.set_task_metadata(task, metadata)

        # Metadata should now be populated
        self.assertDictEqual(metadata, self.wfi.get_task_metadata(task))

    def test_get_task_input(self):
        """Test the obtaining of a task input."""
        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "test_task/output")])
        task = self.wfi.add_task(
            "test_task",
            "ls",
            [StepInput("test_input", "File", "input.txt", "default.txt", "test_input", None, None,
                       None)],
            [StepOutput("test_task/output", "File", "output.txt", "output.txt")])

        self.assertEqual(task.inputs[0], self.wfi.get_task_input(task, "test_input"))

    def test_set_task_input(self):
        """Test the setting of a task input."""
        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "test_task/output")])
        task = self.wfi.add_task(
            "test_task",
            "ls",
            [StepInput("test_input", "File", None, "default.txt", "test_input", None, None, None)],
            [StepOutput("test_task/output", "File", "output.txt", "output.txt")])

        test_input = StepInput("test_input", "File", "input.txt", "default.txt", "test_input",
                               None, None, None)
        self.wfi.set_task_input(task, "test_input", "input.txt")
        self.assertEqual(test_input, self.wfi.get_task_input(task, "test_input"))

    def test_get_task_output(self):
        """Test the obtaining of a task output."""
        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "test_task/output")])
        task = self.wfi.add_task(
            "test_task",
            "ls",
            [StepInput("test_input", "File", "input.txt", "default.txt", "test_input", None, None,
                       None)],
            [StepOutput("test_task/output", "File", "output.txt", "output.txt")])

        self.assertEqual(task.outputs[0], self.wfi.get_task_output(task, "test_task/output"))

    def test_set_task_output(self):
        """Test the setting of a task output."""
        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "test_task/output")])
        task = self.wfi.add_task(
            "test_task",
            "ls",
            [StepInput("test_input", "File", "input.txt", "default.txt", "test_input", None,
                       None, None)],
            [StepOutput("test_task/output", "File", None, "output.txt")])

        test_output = StepOutput("test_task/output", "File", "output.txt", "output.txt")
        self.wfi.set_task_output(task, "test_task/output", "output.txt")
        self.assertEqual(test_output, self.wfi.get_task_output(task, "test_task/output"))

    def test_evaluate_expression(self):
        """Test the evaluation of an input/output expression."""
        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "test_task/output")])
        task = self.wfi.add_task(
            "test_task",
            "ls",
            [StepInput("test_input", "File", "input.txt", "default.txt", "test_input", None, None,
                       '$("test_" + inputs.test_input)')],
            [StepOutput("test_task/output", "File", None, "$(inputs.test_input).bak")])

        test_input = StepInput("test_input", "string", "test_input.txt", "default.txt",
                               "test_input", None, None, '$("test_" + inputs.test_input)')
        self.wfi.evaluate_expression(task, "test_input")
        self.assertEqual(test_input, self.wfi.get_task_input(task, "test_input"))

        test_output = StepOutput("test_task/output", "File", None, "test_input.txt.bak")
        self.wfi.evaluate_expression(task, "test_task/output", output=True)
        self.assertEqual(test_output, self.wfi.get_task_output(task, "test_task/output"))

    def test_workflow_completed(self):
        """Test determining if a workflow has completed."""
        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "test_task/output")])
        task = self.wfi.add_task(
            "test_task",
            "ls",
            [StepInput("test_input", "File", "input.txt", "default.txt", "test_input", None,
                       None, None)],
            [StepOutput("test_task/output", "File", "output.txt",
                        "output.txt")])

        # Workflow not completed
        self.assertFalse(self.wfi.workflow_completed())

        # Not using finalize_task() to avoid unnecessary queries
        self.wfi.set_task_state(task, 'COMPLETED')

        # Workflow now completed
        self.assertTrue(self.wfi.workflow_completed())

    def test_workflow_initialized(self):
        """Test determining if a workflow is initialized."""
        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "viz/output")])

        # Workflow now initialized
        self.assertTrue(self.wfi.workflow_initialized())

    def test_workflow_loaded(self):
        """Test determining if a workflow is loaded."""
        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "viz/output")])
        self.wfi.finalize_workflow()

        # Workflow not loaded
        self.assertFalse(self.wfi.workflow_loaded())

        self.wfi.initialize_workflow(
            generate_workflow_id(), "test_workflow",
            [InputParameter("test_input", "File", "input.txt")],
            [OutputParameter("test_output", "File", "output.txt", "viz/output")])

        # Workflow now loaded
        self.assertTrue(self.wfi.workflow_loaded())

    def test_workflow_id(self):
        """Test retrieving the workflow ID from the workflow interface."""
        self.assertEqual(self.wfi.workflow_id, self.wfi._workflow_id)

        # Set workflow_id to None so it's retrieved from database
        self.wfi._workflow_id = None

        self.assertEqual(self.wfi.workflow_id, self.wfi._workflow_id)

    def _create_test_tasks(self):
        """Create test tasks to reduce redundancy."""
        # Remember that add_task uploads the task to the database as well as returns a Task
        tasks = [
            self.wfi.add_task(
                "data_prep", base_command=["ls", "-a", "-F"],
                inputs=[StepInput("test_input", "File", None, None, "test_input", "-l", None,
                                  None)],
                outputs=[StepOutput("prep/prep_output", "stdout", None, "prep_output.txt")],
                requirements=[Requirement("NetworkAccess", {"networkAccess": True})],
                hints=[Hint("ResourceRequirement", {"ramMax": 2048})], stdout="prep_output.txt"),
            self.wfi.add_task(
                "compute0", base_command="touch",
                inputs=[StepInput("input_data", "File", None, None, "prep/prep_output", None,
                                  None, None)],
                outputs=[StepOutput("compute0/output", "stdout", None, "output0.txt")],
                requirements=[Requirement("NetworkAccess", {"networkAccess": True})],
                hints=[Hint("ResourceRequirement", {"ramMax": 2048})], stdout="output0.txt"),
            self.wfi.add_task(
                "compute1", base_command="find",
                inputs=[StepInput("input_data", "File", None, None, "prep/prep_output", None,
                                  None, None)],
                outputs=[StepOutput("compute1/output", "stdout", None, "output1.txt")],
                requirements=[Requirement("NetworkAccess", {"networkAccess": True})],
                hints=[Hint("ResourceRequirement", {"ramMax": 2048})], stdout="output1.txt"),
            self.wfi.add_task(
                "compute2", base_command="grep",
                inputs=[StepInput("input_data", "File", None, None, "prep/prep_output", None,
                                  None, None)],
                outputs=[StepOutput("compute2/output", "stdout", None, "output2.txt")],
                requirements=[Requirement("NetworkAccess", {"networkAccess": True})],
                hints=[Hint("ResourceRequirement", {"ramMax": 2048})], stdout="output2.txt"),
            self.wfi.add_task(
                "visualization", base_command="python",
                inputs=[StepInput("input0", "File", None, None, "compute0/output", "-i", 1, None),
                        StepInput("input1", "File", None, None, "compute1/output", "-i", 2, None),
                        StepInput("input2", "File", None, None, "compute2/output", "-i", 3, None)],
                outputs=[StepOutput("viz/output", "stdout", "viz_output.txt", "viz_output.txt")],
                requirements=[Requirement("NetworkAccess", {"networkAccess": True})],
                hints=[Hint("ResourceRequirement", {"ramMax": 2048})], stdout="viz_output.txt")
        ]
        return tasks


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['beeflow',
 'beeflow.client',
 'beeflow.common',
 'beeflow.common.build',
 'beeflow.common.cloud',
 'beeflow.common.crt',
 'beeflow.common.db',
 'beeflow.common.gdb',
 'beeflow.common.parser',
 'beeflow.common.worker',
 'beeflow.scheduler',
 'beeflow.tests',
 'beeflow.wf_manager',
 'beeflow.wf_manager.common',
 'beeflow.wf_manager.resources']

package_data = \
{'': ['*'],
 'beeflow': ['data/cloud_templates/*',
             'data/cwl/*',
             'data/cwl/bee_workflows/*',
             'data/cwl/bee_workflows/blast/*',
             'data/cwl/bee_workflows/blast/input/*',
             'data/cwl/bee_workflows/clamr-ci/*',
             'data/cwl/bee_workflows/clamr-wf-checkpoint/*',
             'data/cwl/bee_workflows/clamr-wf-chicoma/*',
             'data/cwl/bee_workflows/clamr-wf-noyaml/*',
             'data/cwl/bee_workflows/clamr-wf-noyaml/lsf-charliecloud/*',
             'data/cwl/bee_workflows/clamr-wf-noyaml/slurm-charliecloud/*',
             'data/cwl/bee_workflows/clamr-wf-noyaml/slurm-singularity/*',
             'data/cwl/bee_workflows/clamr-wf-singularity/*',
             'data/cwl/bee_workflows/clamr-wf-summit/*',
             'data/cwl/bee_workflows/clamr-wf-use-container/*',
             'data/cwl/bee_workflows/clamr-wf/*',
             'data/cwl/bee_workflows/comd-mpi/*',
             'data/cwl/bee_workflows/lulesh-mpi/*',
             'data/cwl/bee_workflows/nwchem-mpi/*',
             'data/cwl/bee_workflows/simple-workflows/*',
             'data/cwl/bee_workflows/simple-workflows/grep-wordcount/*',
             'data/cwl/cwl_validation/*',
             'data/cwl/cwl_validation/builder/*',
             'data/cwl/cwl_validation/grep-wordcount/*',
             'data/dockerfiles/*',
             'data/dockerfiles/comd-pmix-support/*',
             'data/job_templates/*'],
 'beeflow.tests': ['clamr-wf/*']}

install_requires = \
['APScheduler>=3.6.3,<4.0.0',
 'Flask>=2.0,<3.0',
 'Jinja2>=3.0,<4.0',
 'PyYAML>=5.1.1,<6.0.0',
 'cwl-utils>=0.16,<0.17',
 'flask_restful==0.3.9',
 'gunicorn>=20.1.0,<21.0.0',
 'jsonpickle>=2.2.0,<3.0.0',
 'neo4j>=1.7.4,<2.0.0',
 'python-daemon>=2.3.1,<3.0.0',
 'requests-unixsocket>=0.3.0,<0.4.0',
 'typer>=0.5.0,<0.6.0']

extras_require = \
{'cloud_extras': ['google-api-python-client>=2.66.0,<3.0.0',
                  'python-openstackclient>=6.0.0,<7.0.0',
                  'python-heatclient>=3.1.0,<4.0.0']}

entry_points = \
{'console_scripts': ['beecfg = beeflow.common.config_driver:main',
                     'beeclient = beeflow.client.bee_client:main',
                     'beecloud = beeflow.cloud_launcher:main',
                     'beeflow = beeflow.cli:main']}

setup_kwargs = {
    'name': 'hpc-beeflow',
    'version': '0.1.2',
    'description': 'A software package for containerizing HPC applications and managing job workflow',
    'long_description': 'BEE: Build and Execution Environment\n************************************\n\nBEE is a workflow orchestration system designed to build containerized HPC applications and orchestrate workflows across HPC and cloud systems. BEE has adopted the Common Workflow Language (`CWL <https://www.commonwl.org/>`_) for specifying workflows. Complex scientific workflows specified by CWL are managed and visualized through a graph database, giving the user the ability to monitor the state of each task in the workflow. BEE runs jobs using the workload scheduler (i.e. Slurm or LSF) on the HPC system that tasks are specified to run on.\n\nBEE workflows can be archived for provenance and reproducibility. BEE can orchestrate workflows with containerized applications or those built locally on a system. However, there are advantages to containerizing an application.\n\nA container is a package of code (usually binaries) and all of that code\'s dependencies (libraries, etc.). Once built, this container can be run on many different platforms.\n\nContainers provide many benefits:\n\n* Users can choose their own software stack (libraries, compilers, etc.) and not be bound by the currently installed environment on any one machine.\n\n* Codes can be run portably across numerous platforms--all dependencies will be downloaded and installed at run time.\n\n* Entire **workflow** environments can be built into one or more containers. A user can include visualization and analysis tools along with the application. They will all work together as the application runs.\n\n* Provenance and history can be tracked by storing containers in a historical repository. At any time, an older container can be rerun (all of its dependencies are stored with it). Execution is repeatable and interactions between software components can be tracked.\n\n* Functional testing can be performed on smaller, dissimilar machines--there is no real need to test on the actual HPC platform (performance testing obviously requires target hardware).\n\n\nBEE Sites\n=========\n\n* Documentation: `https://lanl.github.io/BEE/ <https://lanl.github.io/BEE/>`_\n\n* Github: `https://github.com/lanl/BEE <https://github.com/lanl/BEE>`_\n\n\nContact\n=======\n\n\nFor bugs and problems report, suggestions and other general questions regarding the BEE project, email questions to `bee-dev@lanl.gov <bee-dev@lanl.gov>`_.\n\n\nContributing\n==========================\n\nThe BEE project adheres to style guidelines specified in `setup.cfg <https://github.com/lanl/BEE/blob/master/setup\\.cfg>`_. Before attempting to commit and push changes, please install our pre-commit githooks by running the following command in project root:\n\nIf using `git --version` >= 2.9:\n    git config core.hooksPath .githooks\n\nOtherwise:\n    cp .githooks/* .git/hooks/\n\nUsing these git hooks will ensure your contributions adhere to style guidelines required for contribution. You will need to repeat these steps for every `BEE` repo you clone.\n\n\nRelease\n==========================\n\nThis software has been approved for open source release and has been assigned **BEE C17056**.\n\nCopyright\n==========================\nLicense can be found `here <https://github.com/lanl/BEE/blob/master/LICENSE>`_\n\n\nPublications\n==========================\n\n- BEE Orchestrator: Running Complex Scientific Workflows on Multiple Systems, HiPC, 2021, `DOI: 10.1109/HiPC53243.2021.00052 <https://doi.org/10.1109/HiPC53243.2021.00052>`_\n- "BeeSwarm: Enabling Parallel Scaling Performance Measurement in Continuous Integration for HPC Applications", ASE, 2021, `DOI: 10.1109/ASE51524.2021.9678805 <https://www.computer.org/csdl/proceedings-article/ase/2021/033700b136/1AjTjgnW2pa#:~:text=10.1109/ASE51524.2021.9678805>`_\n- "BeeFlow: A Workflow Management System for In Situ Processing across HPC and Cloud Systems", ICDCS, 2018, `DOI: 10.1109/ICDCS.2018.00103 <https://ieeexplore.ieee.org/abstract/document/8416366>`_\n- "Build and execution environment (BEE): an encapsulated environment enabling HPC applications running everywhere", IEEE BigData, 2018, `DOI: 10.1109/BigData.2018.8622572 <https://ieeexplore.ieee.org/document/8622572>`_\n\n\n',
    'author': 'BEE-LANL Dev Team',
    'author_email': 'bee-dev@lanl.gov',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/lanl/BEE',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8.3,<=3.11',
}


setup(**setup_kwargs)



"""Tests of the scheduler database."""
import tempfile
import os

import pytest

from beeflow.common.db import sched


@pytest.fixture
def temp_db():
    """Create a fixture for making a temporary datbase."""
    fname = tempfile.mktemp()
    with sched.open_db(fname) as db:
        yield db
    os.remove(fname)


def test_empty(temp_db):
    """Test the empty database."""
    db = temp_db

    assert len(list(db.resources)) == 0


def test_extend(temp_db):
    """Test setting the resources."""
    db = temp_db

    db.resources.extend([1, 2, 3])
    assert list(db.resources) == [1, 2, 3]


def test_clear(temp_db):
    """Test clearing the resources."""
    db = temp_db
    db.resources.extend([8, 9, 10, 11, 12, 13, 14])

    db.resources.clear()
    assert len(list(db.resources)) == 0
# pylama:ignore=W0621



"""BEE configuration driver module."""

from configparser import ConfigParser
import getpass
import os
import platform
import random
import shutil
import sys
import textwrap
import typer

from beeflow.common.config_validator import ConfigValidator
from beeflow.common.cli import NaturalOrderGroup


# System specific path set up
HOME_DIR = os.path.expanduser('~/')
_SYSTEM = platform.system()
if _SYSTEM == 'Linux':
    CONF_DIR = os.path.join(HOME_DIR, '.config/beeflow')
    SYSCONFIG_FILE = '/etc/beeflow/bee.conf'
    USERCONFIG_FILE = os.path.expanduser('~/.config/beeflow/bee.conf')
elif _SYSTEM == 'Darwin':
    CONF_DIR = os.path.expanduser('~/Library/Application Support/beeflow')
    SYSCONFIG_FILE = '/Library/Application Support/beeflow/bee.conf'
    USERCONFIG_FILE = os.path.expanduser(
        '~/Library/Application Support/beeflow/bee.conf')
elif _SYSTEM == 'Windows':
    CONF_DIR = os.path.expandvars(r'%APPDATA%\beeflow')
    SYSCONFIG_FILE = None
    USERCONFIG_FILE = os.path.expandvars(r'%APPDATA%\beeflow\bee.conf')
else:
    raise RuntimeError(f'System "{_SYSTEM}" is not supported')


class BeeConfig:
    r"""Class to manage and store all BEE configuration.

    All configuration values can be retrieved by using the get(section, key)
    class method. If those particular values are not in the config file, then
    defaults will be set, or a validation error raised, based on the validation code
    within the module.

    When new configuration keys are needed, they should be added to the
    validation code, along with information about the allowed input types, and
    general text for the user. If a key isn't added, but other code tries to
    access it, then an error will be raised letting the programmer know that the
    key needs to be added to the validation. This will hopefully help manage
    complexity and act as documentation as more keys are added.

    Configuration file locations by supported platform:

    Linux:
      sysconfig_file = '/etc/beeflow/bee.conf'

      userconfig_file = '~/.config/beeflow/bee.conf'

    MacOS:
      sysconfig_file = '/Library/Application Support/beeflow/bee.conf'

      userconfig_file = '~/Library/Application Support/beeflow/bee.conf'

    Windows:
      sysconfig_file = NOT SUPPORTED. Should be windows registry.

      userconfig_file = '%APPDATA%\\beeflow\\bee.conf'
    """

    CONFIG = None

    def __init__(self, **kwargs):
        """Do not use this constructor."""
        raise RuntimeError(
            'BeeConfig is a singleton class. Call BeeConfig.init() once to initialize.'
        )

    @classmethod
    def ready(cls):
        """Check if the class has been initialized."""
        return cls.CONFIG is not None

    @classmethod
    def init(cls, userconfig=None, **_kwargs):
        """Initialize BeeConfig class.

        We check the platform and read in system and user configuration files.
        If the user configuration file doesn't exist we create it with a [DEFAULT] section.
        """
        global USERCONFIG_FILE
        if cls.CONFIG is not None:
            return
        config = ConfigParser()
        if userconfig is not None:
            USERCONFIG_FILE = userconfig
        # Try and read the file
        try:
            with open(USERCONFIG_FILE, encoding='utf-8') as fp:
                config.read_file(fp)
        except FileNotFoundError:
            sys.exit('Configuration file does not exist! Please try running `beecfg new`.')
        # remove default keys from the other sections
        default_keys = list(config['DEFAULT'])
        config = {sec_name: {key: config[sec_name][key] for key in config[sec_name]
                             if sec_name == 'DEFAULT' or key not in default_keys} # noqa
                  for sec_name in config}
        # Validate the config
        cls.CONFIG = VALIDATOR.validate(config)

    @classmethod
    def userconfig_path(cls):
        """Get the path of the user config."""
        return USERCONFIG_FILE

    @classmethod
    def get(cls, sec_name, opt_name):
        """Get a configuration value.

        If this throws, then either BeeConfig has not been initialized or a
        configuration value is missing from the definition. Default values
        are built into the ConfigValidator class, so there is no need to
        specify a default or a fallback here.
        """
        if cls.CONFIG is None:
            raise RuntimeError('BeeConfig has not been initialized')
        try:
            return cls.CONFIG[sec_name][opt_name] # noqa (this object is subscritable)
        except KeyError:
            raise RuntimeError(
                f'Option {sec_name}::{opt_name} was not found. Please contact '
                'a BEE developer or check the validation code in '
                'beeflow/common/config_driver.py.'
            ) from None

    @staticmethod
    def resolve_path(relative_path):
        """Resolve relative paths to absolute paths.

        :param relative_path: Input path. May include "../"
        :type relative_path: string, path to file
        """
        # Discard redundant paths, iff Windows, replace(\,/)
        # Assume "~" means home dir
        relative_path = os.path.expanduser(os.path.normpath(relative_path))
        # Resolve the true path (expand relative path refs)
        tmp = os.getcwd()
        if os.path.isdir(relative_path):
            os.chdir(relative_path)
            absolute_path = os.getcwd()
        else:
            # Get desired config file name
            filename = os.path.basename(relative_path)
            os.chdir(os.path.dirname(relative_path))
            absolute_path = '/'.join([os.getcwd(), filename])
        os.chdir(tmp)
        return absolute_path


def check_yes(msg):
    """Check for a y/n answer."""
    res = input(f'{msg} [y/n] ')
    return res.lower() == 'y'


# Specialized functions for validation and config initialization

def validate_path(path):
    """Check that the path exists."""
    path = os.path.expanduser(path)
    if not os.path.exists(path):
        raise ValueError(f'path "{path}" does not exist')
    return path


def validate_dir(path):
    """Check that the path exists and is a directory."""
    path = validate_path(path)
    if not os.path.isdir(path):
        raise ValueError('path "{path}" is not a directory')
    return path


def validate_make_dir(path):
    """Check if the dir exists and if not create it."""
    os.makedirs(path, exist_ok=True)
    if not os.path.isdir(path):
        raise ValueError('path "{path}" is not a directory')
    return path


def validate_file(path):
    """Check that the path exists and is a file."""
    path = validate_path(path)
    if not os.path.isfile(path):
        raise ValueError(f'path "{path}" is not a file')
    return path


def validate_nonnegative_int(value):
    """Validate that the input is nonnegative."""
    i = int(value)
    if i < 0:
        raise ValueError('the value must be nonnegative')
    return i


def validate_bool(value):
    """Validate a boolean value."""
    return str(value).lower() == 'true'


def check_choice(msg, opts):
    """Ask the user to pick from opts."""
    while True:
        value = input(f'{msg} ({",".join(opts)})')
        if value in opts:
            return value
        print(f'ERROR: invalid option: {value}')


def join_path(*pargs):
    """Join multiple dirnames together to form a path."""
    path = pargs[0]
    for part in pargs[1:]:
        path = os.path.join(path, part)
    return path


def job_template_init(path, cur_opts):
    """Job template init function.

    :param path: chosen path of Jinja job template file
    :param cur_opts: current chosen options from the config generator
    :return: initialized Jinja template path
    """
    path = os.path.expanduser(path)
    if os.path.exists(path):
        return path
    print('Check that options (like accounts)for your particular system are satisfied.')
    if not check_yes(f'Do you want to write a default job template to:\n "{path}"?'):
        return path
    # Check for workload_scheduler in cur_opts (NB: this will need to be updated
    # if the option name changes later on)
    if 'DEFAULT' in cur_opts and 'workload_scheduler' in cur_opts['DEFAULT']:
        template = cur_opts['DEFAULT']['workload_scheduler']
    else:
        template = check_choice('What template should be generated?', ('Slurm', 'LSF', 'Simple'))
    template_files = {
        'Slurm': 'slurm-submit.jinja',
        'LSF': 'lsf-submit.jinja',
    }
    if template not in template_files:
        raise NotImplementedError(f'generation of template file "{template}" is not implemented')
    fname = template_files[template]
    # I don't like how this is done, but there seems to be some difficulties
    # with using the pkgutil.get_data() method and the importlib.resources
    # module.
    common = os.path.dirname(__file__)
    beeflow = os.path.dirname(common)
    file_path = join_path(beeflow, 'data', 'job_templates', fname)
    shutil.copy(file_path, path)
    return path


def bee_workdir_init(path, _cur_opts):
    """BEE workdir init function.

    :param path: chosen path for the bee workdir
    :param _cur_opts: current chosen options form the config generator
    :return: initialized bee workdir path
    """
    path = os.path.expanduser(path)
    if os.path.exists(path):
        return path
    if not check_yes(f'Path "{path}" does not exist.\nWould you like to create it?'):
        return path
    os.makedirs(path)
    return path


# Below is the definition of all bee config options, defaults and requirements.
# This will be used to validate config files on loading them in the BeeConfig
# singleton class above.
#
# Set up default ports, get parent's pid to offset ports.
# Windows note: uid method better but not available in Windows
if platform.system() == 'Windows':
    # Would prefer something like a uid for windows.
    OFFSET = os.getppid() % 100
else:
    OFFSET = os.getuid() % 100
DEFAULT_BOLT_PORT = 7687 + OFFSET
DEFAULT_HTTP_PORT = 7474 + OFFSET
DEFAULT_HTTPS_PORT = 7473 + OFFSET

DEFAULT_WFM_PORT = 5000 + OFFSET
DEFAULT_TM_PORT = 5050 + OFFSET
DEFAULT_SCHED_PORT = 5100 + OFFSET

SOCKET_PATH = join_path(HOME_DIR, '.beeflow', 'sockets')
os.makedirs(SOCKET_PATH, exist_ok=True)
DEFAULT_WFM_SOCKET = join_path(SOCKET_PATH, 'wf_manager.sock')
DEFAULT_TM_SOCKET = join_path(SOCKET_PATH, 'task_manager.sock')
DEFAULT_SCHED_SOCKET = join_path(SOCKET_PATH, 'scheduler.sock')
DEFAULT_BEEFLOW_SOCKET = join_path(SOCKET_PATH, 'beeflow.sock')

DEFAULT_BEE_WORKDIR = join_path(HOME_DIR, '.beeflow')
USER = getpass.getuser()
# Create the validator
VALIDATOR = ConfigValidator('BEE configuration file and validation information.')
VALIDATOR.section('DEFAULT', info='Default bee.conf configuration section.')
VALIDATOR.option('DEFAULT', 'bee_workdir', info='main BEE workdir',
                 attrs={'default': DEFAULT_BEE_WORKDIR}, validator=validate_make_dir)
VALIDATOR.option('DEFAULT', 'workload_scheduler', choices=('Slurm', 'LSF', 'Simple'),
                 info='backend workload scheduler to interact with ')
VALIDATOR.option('DEFAULT', 'use_archive', validator=validate_bool, attrs={'default': True},
                 info='use the BEE archiving functinality')
VALIDATOR.option('DEFAULT', 'bee_dep_image', validator=validate_file,
                 info='container image with BEE dependencies')
VALIDATOR.option('DEFAULT', 'beeflow_pidfile',
                 attrs={'default': join_path(DEFAULT_BEE_WORKDIR, 'beeflow.pid')},
                 info='location of beeflow pidfile')
VALIDATOR.option('DEFAULT', 'beeflow_socket',
                 attrs={'default': DEFAULT_BEEFLOW_SOCKET},
                 info='location of beeflow socket')
# Workflow Manager
VALIDATOR.section('workflow_manager', info='Workflow manager section.')
VALIDATOR.option('workflow_manager', 'socket',
                 attrs={'default': DEFAULT_WFM_SOCKET}, validator=str,
                 info='workflow manager port')
# Task manager
VALIDATOR.section('task_manager',
                  info='Task manager configuration and config of container to use.')
VALIDATOR.option('task_manager', 'socket', attrs={'default': DEFAULT_TM_SOCKET}, validator=str,
                 info='task manager listen port')
VALIDATOR.option('task_manager', 'container_runtime', attrs={'default': 'Charliecloud'},
                 choices=('Charliecloud', 'Singularity'),
                 info='container runtime to use for configuration')
VALIDATOR.option('task_manager', 'runner_opts', validator=str, attrs={'default': ''},
                 info='special runner options to pass to the runner opts')
# Note: I've added a special attrs keyword which can include anything. In this
# case it's being used to store a special 'init' function that can be used to
# initialize the parameter when a user is generating a new configuration.
VALIDATOR.option('task_manager', 'job_template',
                 info='job template to use for generating submission scripts',
                 validator=validate_file,
                 attrs={'init': job_template_init, 'default': join_path(CONF_DIR, 'submit.jinja')})
# Charliecloud (depends on task_manager::container_runtime == Charliecloud)
VALIDATOR.section('charliecloud', info='Charliecloud configuration section.',
                  depends_on=('task_manager', 'container_runtime', 'Charliecloud'))
VALIDATOR.option('charliecloud', 'image_mntdir', attrs={'default': join_path('/tmp', USER)},
                 info='Charliecloud mount directory', validator=validate_make_dir)


def validate_chrun_opts(opts):
    """Ensure that chrun_opts don't contain options that'll conflict with BEE."""
    args = opts.split()
    if '--cd' in args or '-c' in args:
        raise ValueError('Initial working directory option (-c or --cd) conflicts with BEE. '
                         'Please make sure that you choose the correct working directory on '
                         'workflow submission.')
    return opts


VALIDATOR.option('charliecloud', 'chrun_opts', attrs={'default': ''},
                 validator=validate_chrun_opts,
                 info='extra options to pass to ch-run')
VALIDATOR.option('charliecloud', 'setup', attrs={'default': ''}, validator=str,
                 info='extra Charliecloud setup to put in a job script')
# Graph Database
VALIDATOR.section('graphdb', info='Main graph database configuration section.')
VALIDATOR.option('graphdb', 'hostname', attrs={'default': 'localhost'},
                 info='hostname of database')
VALIDATOR.option('graphdb', 'dbpass', attrs={'default': 'password'}, info='password for database')
VALIDATOR.option('graphdb', 'bolt_port', attrs={'default': DEFAULT_BOLT_PORT}, validator=int,
                 info='port used for the BOLT API')
VALIDATOR.option('graphdb', 'http_port', attrs={'default': DEFAULT_HTTP_PORT}, validator=int,
                 info='HTTP port used for the graph database')
VALIDATOR.option('graphdb', 'https_port', attrs={'default': DEFAULT_HTTPS_PORT},
                 info='HTTPS port used for the graph database')
VALIDATOR.option('graphdb', 'gdb_image_mntdir', attrs={'default': join_path('/tmp', USER)},
                 info='graph database image mount directory', validator=validate_make_dir)
VALIDATOR.option('graphdb', 'sleep_time', validator=int, attrs={'default': 10},
                 info='how long to wait for the graph database to come up (this can take a while, '
                      'depending on the system)')
# Builder
VALIDATOR.section('builder', info='General builder configuration section.')
VALIDATOR.option('builder', 'deployed_image_root', attrs={'default': '/tmp'},
                 info='where to deploy container images', validator=validate_make_dir)
VALIDATOR.option('builder', 'container_output_path', attrs={'default': '/tmp'},
                 info='container output path', validator=validate_make_dir)
VALIDATOR.option('builder', 'container_archive',
                 attrs={'default': join_path(DEFAULT_BEE_WORKDIR, 'container_archive')},
                 info='container archive location')
VALIDATOR.option('builder', 'container_type', attrs={'default': 'charliecloud'},
                 info='container type to use')
# Slurmrestd (depends on DEFAULT:workload_scheduler == Slurm)
VALIDATOR.section('slurmrestd', info='Configuration section for Slurmrestd.',
                  depends_on=('DEFAULT', 'workload_scheduler', 'Slurm'))
DEFAULT_SLURMRESTD_SOCK = join_path('/tmp', f'slurm_{USER}_{random.randint(1, 10000)}.sock')
VALIDATOR.option('slurmrestd', 'slurm_socket', attrs={'default': DEFAULT_SLURMRESTD_SOCK},
                 info='socket location')
VALIDATOR.option('slurmrestd', 'slurm_args', attrs={'default': '-s openapi/v0.0.35'},
                 info='arguments for the slurmrestd binary')
# Scheduler
VALIDATOR.section('scheduler', info='Scheduler configuration section.')
VALIDATOR.option('scheduler', 'log',
                 attrs={'default': join_path(DEFAULT_BEE_WORKDIR, 'logs', 'scheduler.log')},
                 validator=str, info='scheduler log file')
VALIDATOR.option('scheduler', 'socket', attrs={'default': DEFAULT_SCHED_SOCKET}, validator=str,
                 info='scheduler socket')
VALIDATOR.option('scheduler', 'alloc_logfile',
                 attrs={'default': join_path(DEFAULT_BEE_WORKDIR, 'logs', 'scheduler_alloc.log')},
                 validator=str, info='allocation logfile, to be used for later training')
SCHEDULER_ALGORITHMS = ('fcfs', 'backfill', 'sjf')
VALIDATOR.option('scheduler', 'algorithm', attrs={'default': 'fcfs'}, choices=SCHEDULER_ALGORITHMS,
                 info='scheduling algorithm to use')
VALIDATOR.option('scheduler', 'default_algorithm', attrs={'default': 'fcfs'},
                 choices=SCHEDULER_ALGORITHMS,
                 info=('default algorithm to use'))
VALIDATOR.option('scheduler', 'workdir',
                 attrs={'default': join_path(DEFAULT_BEE_WORKDIR, 'scheduler')},
                 validator=str,
                 info='workdir to be used for the scheduler')


def print_wrap(text, next_line_indent=''):
    """Print while wrapping lines to make the output easier to read."""
    for line in textwrap.wrap(text, width=80, subsequent_indent=next_line_indent):
        print(line)


class ConfigGenerator:
    """Config generator class."""

    def __init__(self, fname, validator):
        """Construct the config generator."""
        self.fname = fname
        self.validator = validator
        self.sections = {}

    def choose_values(self):
        """Choose configuration values based on user input."""
        dirname = os.path.dirname(self.fname)
        if dirname:
            os.makedirs(dirname, exist_ok=True)
        print(f'Creating a new config: "{self.fname}".')
        print()
        print_wrap('This will walk you through creating a new configuration for BEE. '
                   'Note that you will only be required to enter values for options '
                   'without defaults. Please take a look at the other options and '
                   'their defaults before running BEE.')
        print()
        print('Please enter values for the following sections and options:')
        # Let the user choose values for each required attribute
        for sec_name, section in self.validator.sections:
            # Determine if this section is valid under the current configuration
            if not self.validator.is_section_valid(self.sections, sec_name):
                continue
            self.sections[sec_name] = {}
            printed = False
            for opt_name, option in self.validator.options(sec_name):
                # Print the section name if it hasn't already been printed.
                if not printed:
                    print()
                    print(f'## {sec_name}')
                    print()
                    print_wrap(section.info)
                    print()
                    printed = True
                if option.attrs is not None and 'default' in option.attrs:
                    default = option.attrs['default']
                    if option.attrs is not None and 'init' in option.attrs:
                        value = option.attrs['init'](default, self.sections)
                    else:
                        value = option.validate(default)
                    print(f'Setting option "{opt_name}" to default value "{value}".')
                    print()
                    self.sections[sec_name][opt_name] = value
                    continue
                value = None
                # Input and then validate the value
                while value is None:
                    print_wrap(f'{opt_name} - {option.info}')
                    if option.choices is not None:
                        print(f'(allowed values: {",".join(option.choices)})')
                    value = input(f'{opt_name}: ')
                    # Call the init function if there is one
                    if option.attrs is not None and 'init' in option.attrs:
                        option.attrs['init'](value, self.sections)
                    try:
                        option.validate(value)
                    except ValueError as err:
                        print('ERROR:', err)
                        value = None
                print()
                self.sections[sec_name][opt_name] = value
        return self

    def save(self):
        """Save the config to a file."""
        print()
        print('The following configuration options were chosen:')
        for sec_name, section in self.sections.items():
            print()
            print(f'[{sec_name}]')
            for opt_name in section:
                print(f'{opt_name} = {section[opt_name]}')
        print()
        ans = input('Would you like to save this config? [y/n] ')
        if ans.lower() != 'y':
            print('Quitting without saving')
            return
        try:
            with open(self.fname, 'w', encoding='utf-8') as fp:
                print('# BEE Configuration File', file=fp)
                for sec_name, section in self.sections.items():
                    if not section:
                        continue
                    print(file=fp)
                    print(f'[{sec_name}]'.format(sec_name), file=fp)
                    for opt_name in section:
                        print(f'{opt_name} = {section[opt_name]}', file=fp)
        except FileNotFoundError:
            print('Configuration file does not exist!')
        print(70 * '#')
        print('Before running BEE, check defaults in the configuration file:',
              f'\n\t{self.fname}',
              '\n ** See documentation for values you should refrain from editing! **',
              '\n\nThe job template configured is:',
              f'\n\t{self.sections["task_manager"]["job_template"]}',
              '\n ** Include job options (such as account) required for this system.**')
        print('\n(Try `beecfg info` to see more about each option)')
        print(70 * '#')


app = typer.Typer(no_args_is_help=False, add_completion=False, cls=NaturalOrderGroup)


@app.command()
def validate(path: str = typer.Argument(default=USERCONFIG_FILE,
                                        help='Path to config file')):
    """Validate an existing configuration file."""
    BeeConfig.init(path)


@app.command()
def info():
    """Display some info about bee.conf's various options."""
    print('# BEE Configuration')
    print()
    print_wrap(VALIDATOR.description)
    print()
    for sec_name, section in VALIDATOR.sections:
        print(f'## {sec_name}')
        if section.depends_on is not None:
            print()
            deps = section.depends_on
            print_wrap(f'*only required if {deps[0]}::{deps[1]} == "{deps[2]}"*')
        print()
        print_wrap(f'{section.info}')
        print()
        for opt_name, option in VALIDATOR.options(sec_name):
            print_wrap(f'* {opt_name} - {option.info}', '  ')
            if option.choices is not None:
                print(f'\t* allowed values: {",".join(option.choices)}')
        print()


@app.command()
def new(path: str = typer.Argument(default=USERCONFIG_FILE,
                                   help='Path to new config file')):
    """Create a new config file."""
    if os.path.exists(path):
        if check_yes(f'Path "{path}" already exists.\nWould you like to save a copy of it?'):
            i = 1
            backup_path = f'{path}.{i}'
            while os.path.exists(backup_path):
                i += 1
                backup_path = f'{path}.{i}'
            shutil.copy(path, backup_path)
            print(f'Saved old config to "{backup_path}".')
            print()
    ConfigGenerator(path, VALIDATOR).choose_values().save()


@app.command()
def show(path: str = typer.Argument(default=USERCONFIG_FILE,
                                    help='Path to config file')):
    """Show the contents of bee.conf."""
    if not os.path.exists(path):
        print('The bee.conf does not exist yet. Please run `beecfg new`.')
        return
    print(f'# {path}')
    with open(path, encoding='utf-8') as fp:
        print(fp.read(), end='')


def main():
    """Entry point for config validation and help."""
    app()


if __name__ == '__main__':
    app()



"""Mid-level interface for worker, a work load manager.

Delegates the actual work to an instance of a subclass of
the abstract base class 'Worker'. Default: 'SlurmWorker' class.
"""

from beeflow.common.worker.slurm_worker import SlurmWorker
from beeflow.common.worker.lsf_worker import LSFWorker


class WorkerInterface:
    """Interface for monitoring and managing workloads and jobs.

    Requires an implemented subclass of Worker to function.
    """

    def __init__(self, worker=SlurmWorker, **kwargs):
        """Initialize the workload interface with a workload driver, SlurmWorker by default.

        :param worker: the work load driver (SlurmWorker by default)
        :type worker: subclass of Worker
        """
        self._worker = worker(**kwargs)

    def submit_task(self, task):
        """Worker builds script and submits task as job returns job_id, job_state.

        :param task: instance of Task
        :rtype: tuple (int, string)
        """
        # First prepare for the task (create necessary directories, etc.)
        self._worker.prepare(task)
        # Then submit it to the worker
        return self._worker.submit_task(task)

    def cancel_task(self, job_id):
        """Cancel job for task with job_id.

        :param job_id: job_id to be cancelled
        :type job_id: integer
        :rtype: string
        """
        return self._worker.cancel_task(job_id)

    def query_task(self, job_id):
        """Query state of job with job_id; returns job_state.

        :param job_id: job id to query for status.
        :type job_id: int
        :rtype: tuple (int, string)
        """
        return self._worker.query_task(job_id)
# Ignore W0611 module imported but unused error; unsure which workload scheduler will be needed
# pylama:ignore=W0611



"""Path conversion code."""
import os


class PathError(Exception):
    """Path error class."""

    def __init__(self, *args):
        """Construct a path error object."""
        self.args = args


def _components(path):
    """Convert a path into a list of components."""
    if not os.path.isabs(path):
        raise PathError('Bind mounts and workdir paths must be absolute')
    path = os.path.normpath(path)
    return [comp for comp in path.split('/') if comp]


def convert_path(path, bind_mounts):
    """Convert a path outside the container to a path inside the container."""
    comps = _components(path)
    for outside, inside in bind_mounts.items():
        outside = _components(outside)
        inside = _components(inside)
        if comps[:len(outside)] == outside:
            base = comps[len(outside):]
            inside.extend(base)
            new_path = '/'.join(inside)
            return f'/{new_path}'
    return path



"""Singularity driver as the container runtime system for tasks.

Creates text for tasks using Singularity.
"""

from beeflow.common.crt.crt_driver import (ContainerRuntimeDriver, ContainerRuntimeResult)
from beeflow.common.build.build_driver import task2arg


class SingularityDriver(ContainerRuntimeDriver):
    """The ContainerRuntimeDriver for Singularity as container runtime system.

    Creates the text for the task for using Singularity to test abstract class.
    """

    def run_text(self, task):
        """Build text for Singularity batch script."""
        # Make sure all commands are strings
        cmd_tasks = list(map(str, task.command))
        main_command = cmd_tasks
        if task.hints is not None:
            hints = dict(task.hints)
            try:
                img = hints['DockerRequirement']['beeflow:copyContainer']
                argv = ['singularity', 'exec']
                if task.workdir is not None:
                    argv.extend(['--pwd', task.workdir])
                argv.append(img)
                argv.extend(cmd_tasks)
                main_command = argv
            except (KeyError, TypeError):
                pass
        # Change to the working directory
        env_code = ''
        if task.workdir is not None:
            env_code = f'cd {task.workdir}\n'
        return ContainerRuntimeResult(env_code=env_code, pre_commands=[],
                                      main_command=main_command, post_commands=[])

    def build_text(self, userconfig, task):
        """Build text for Singularity batch script."""
        task_args = task2arg(task)
        text = (f'beeflow --build {userconfig} {task_args}\n'
                )
        return text



"""Neo4j interface module.

Connection requires a valid URI, Username, and Password.
The current defaults are defined below, but should later be
either standardized or read from a config file.
"""

from neo4j import GraphDatabase as Neo4jDatabase
from neobolt.exceptions import ServiceUnavailable

from beeflow.common.gdb.gdb_driver import GraphDatabaseDriver
from beeflow.common.gdb import neo4j_cypher as tx
from beeflow.common.wf_data import (Workflow, Task, Requirement, Hint,
                                    InputParameter, OutputParameter, StepInput, StepOutput)

# Default Neo4j authentication
# We may want to instead get these from a config at some point
DEFAULT_HOSTNAME = "localhost"
DEFAULT_BOLT_PORT = "7687"
DEFAULT_USER = "neo4j"
DEFAULT_PASSWORD = "password"


class Neo4JNotRunning(Exception):
    """Exception thrown when connection attempted while Neo4j is not running."""


class Neo4jDriver(GraphDatabaseDriver):
    """The driver for a Neo4j Database.

    Implements GraphDatabaseDriver.
    Wraps the neo4j package proprietary driver.
    """

    def __init__(self, user=DEFAULT_USER, password=DEFAULT_PASSWORD, **kwargs):
        """Create a new Neo4j database driver.

        :param uri: the URI of the Neo4j database
        :type uri: str
        :param user: the username for the database user account
        :type user: str
        :param password: the password for the database user account
        :type password: str
        """
        db_hostname = kwargs.get("db_hostname", DEFAULT_HOSTNAME)
        bolt_port = kwargs.get("bolt_port", DEFAULT_BOLT_PORT)
        password = kwargs.get("db_pass", DEFAULT_PASSWORD)
        uri = f"bolt://{db_hostname}:{bolt_port}"

        try:
            # Connect to the Neo4j database using the Neo4j proprietary driver
            self._driver = Neo4jDatabase.driver(uri, auth=(user, password))
        except ServiceUnavailable as sue:
            raise Neo4JNotRunning("Neo4j database is unavailable") from sue

    def initialize_workflow(self, workflow):
        """Begin construction of a workflow stored in Neo4j.

        Creates the Workflow, Requirement, and Hint nodes in the Neo4j database.

        :param workflow: the workflow description
        :type workflow: Workflow
        """
        with self._driver.session() as session:
            session.write_transaction(tx.create_workflow_node, workflow)
            session.write_transaction(tx.create_workflow_requirement_nodes,
                                      requirements=workflow.requirements)
            session.write_transaction(tx.create_workflow_hint_nodes, hints=workflow.hints)
            session.write_transaction(tx.create_workflow_input_nodes, inputs=workflow.inputs)
            session.write_transaction(tx.create_workflow_output_nodes, outputs=workflow.outputs)

    def execute_workflow(self):
        """Begin execution of the workflow stored in the Neo4j database."""
        self._write_transaction(tx.set_init_task_inputs)
        self._write_transaction(tx.set_init_tasks_to_ready)
        self._write_transaction(tx.set_workflow_state, state='RUNNING')

    def pause_workflow(self):
        """Pause execution of a running workflow in Neo4j.

        Sets tasks with state 'RUNNING' to 'PAUSED'.
        """
        with self._driver.session() as session:
            session.write_transaction(tx.set_workflow_state, state='PAUSED')

    def resume_workflow(self):
        """Resume execution of a paused workflow in Neo4j.

        Sets workflow state to 'PAUSED'
        """
        with self._driver.session() as session:
            session.write_transaction(tx.set_workflow_state, state='RESUME')

    def reset_workflow(self, new_id):
        """Reset the execution state of an entire workflow.

        Sets all task states to 'WAITING'.
        Changes the workflow ID of the Workflow and Task nodes with new_id.

        :param new_id: the new workflow ID
        :type new_id: str
        """
        with self._driver.session() as session:
            session.write_transaction(tx.reset_tasks_metadata)
            session.write_transaction(tx.reset_workflow_id, new_id=new_id)

    def load_task(self, task):
        """Load a task into a workflow stored in the Neo4j database.

        Dependencies are automatically deduced and generated by Neo4j upon loading
        each task by matching task inputs and outputs.

        Task hint nodes and metadata nodes are created for querying convenience.

        :param task: a workflow task
        :type task: Task
        """
        with self._driver.session() as session:
            session.write_transaction(tx.create_task, task=task)
            session.write_transaction(tx.create_task_hint_nodes, task=task)
            session.write_transaction(tx.create_task_requirement_nodes, task=task)
            session.write_transaction(tx.create_task_input_nodes, task=task)
            session.write_transaction(tx.create_task_output_nodes, task=task)
            session.write_transaction(tx.create_task_metadata_node, task=task)
            session.write_transaction(tx.add_dependencies, task=task)

    def initialize_ready_tasks(self):
        """Set runnable tasks to state 'READY'.

        Runnable tasks are tasks with all input dependencies fulfilled.
        """
        self._write_transaction(tx.set_runnable_tasks_to_ready)

    def restart_task(self, old_task, new_task):
        """Restart a failed task.

        Create a Task node for new_task with 'RESTARTED_FROM' relationship to the
        Task node of old_task.

        :param old_task: the failed task
        :type old_task: Task
        :param new_task: the new (restarted) task
        :type new_task: Task
        """
        with self._driver.session() as session:
            session.write_transaction(tx.create_task, task=new_task)
            session.write_transaction(tx.create_task_hint_nodes, task=new_task)
            session.write_transaction(tx.create_task_requirement_nodes, task=new_task)
            session.write_transaction(tx.create_task_input_nodes, task=new_task)
            session.write_transaction(tx.create_task_output_nodes, task=new_task)
            session.write_transaction(tx.create_task_metadata_node, task=new_task)
            session.write_transaction(tx.add_dependencies, task=new_task, old_task=old_task,
                                      restarted_task=True)

    def finalize_task(self, task):
        """Set task state to 'COMPLETED' and set inputs from source.

        :param task: the task to finalize
        :type task: Task
        """
        self._write_transaction(tx.set_task_state, task=task, state="COMPLETED")
        self._write_transaction(tx.copy_task_outputs, task=task)

    def get_task_by_id(self, task_id):
        """Return a reconstructed task from the Neo4j database.

        :param task_id: a task's ID
        :type task_id: str
        :rtype: Task
        """
        task_record = self._read_transaction(tx.get_task_by_id, task_id=task_id)
        tuples = self._get_task_data_tuples([task_record])
        return _reconstruct_task(tuples[0][0], tuples[0][1], tuples[0][2], tuples[0][3],
                                 tuples[0][4])

    def get_workflow_description(self):
        """Return a reconstructed Workflow object from the Neo4j database.

        :rtype: Workflow
        """
        workflow_record = self._read_transaction(tx.get_workflow_description)
        requirements, hints = self.get_workflow_requirements_and_hints()
        inputs, outputs = self.get_workflow_inputs_and_outputs()
        return _reconstruct_workflow(workflow_record, hints, requirements, inputs, outputs)

    def get_workflow_state(self):
        """Return the current workflow state from the Neo4j database.

        :rtype: str
        """
        return self._read_transaction(tx.get_workflow_state)

    def set_workflow_state(self, state):
        """Set the state of the workflow in the Neo4j database.

        :param state: the new state of the workflow
        :type state: str
        """
        self._write_transaction(tx.set_workflow_state, state=state)

    def get_workflow_tasks(self):
        """Return all workflow task records from the Neo4j database.

        :rtype: list of Task
        """
        task_records = self._read_transaction(tx.get_workflow_tasks)
        tuples = self._get_task_data_tuples(task_records)
        return [_reconstruct_task(tup[0], tup[1], tup[2], tup[3], tup[4]) for tup in tuples]

    def get_workflow_requirements_and_hints(self):
        """Return all workflow requirements and hints from the Neo4j database.

        Returns a tuple of (requirements, hints).

        :rtype: (list of Requirement, list of Hint)
        """
        with self._driver.session() as session:
            requirements = _reconstruct_requirements(
                session.read_transaction(tx.get_workflow_requirements))
            hints = _reconstruct_hints(session.read_transaction(tx.get_workflow_hints))
        return requirements, hints

    def get_workflow_inputs_and_outputs(self):
        """Return all workflow inputs and outputs from the Neo4j database.

        Returns a tuple of (inputs, outputs).

        :rtype: (list of InputParameter, list of OutputParameter)
        """
        with self._driver.session() as session:
            inputs = _reconstruct_workflow_inputs(session.read_transaction(tx.get_workflow_inputs))
            outputs = _reconstruct_workflow_outputs(
                session.read_transaction(tx.get_workflow_outputs))

        return inputs, outputs

    def get_ready_tasks(self):
        """Return tasks with state 'READY' from the graph database.

        :rtype: list of Task
        """
        task_records = self._read_transaction(tx.get_ready_tasks)
        tuples = self._get_task_data_tuples(task_records)
        return [_reconstruct_task(tup[0], tup[1], tup[2], tup[3], tup[4]) for tup in tuples]

    def get_dependent_tasks(self, task):
        """Return the dependent tasks of a specified workflow task.

        :param task: the task whose dependents to retrieve
        :type task: Task
        :rtype: list of Task
        """
        task_records = self._read_transaction(tx.get_dependent_tasks, task=task)
        tuples = self._get_task_data_tuples(task_records)
        return [_reconstruct_task(tup[0], tup[1], tup[2], tup[3], tup[4]) for tup in tuples]

    def get_task_state(self, task):
        """Return the state of a task in the Neo4j workflow.

        :param task: the task whose state to retrieve
        :type task: Task
        :rtype: str
        """
        return self._read_transaction(tx.get_task_state, task=task)

    def set_task_state(self, task, state):
        """Set the state of a task in the Neo4j workflow.

        :param task: the task whose state to change
        :type task: Task
        :param state: the new state
        :type state: str
        """
        self._write_transaction(tx.set_task_state, task=task, state=state)

    def get_task_metadata(self, task):
        """Return the metadata of a task in the Neo4j workflow.

        :param task: the task whose metadata to retrieve
        :type task: Task
        :rtype: dict
        """
        metadata_record = self._read_transaction(tx.get_task_metadata, task=task)
        return _reconstruct_metadata(metadata_record)

    def set_task_metadata(self, task, metadata):
        """Set the metadata of a task in the Neo4j workflow.

        :param task: the task whose metadata to set
        :type task: Task
        :param metadata: the job description metadata
        :type metadata: dict
        """
        self._write_transaction(tx.set_task_metadata, task=task, metadata=metadata)

    def get_task_input(self, task, input_id):
        """Get a task input object.

        :param task: the task whose input to retrieve
        :type task: Task
        :param input_id: the ID of the input
        :type input_id: str
        :rtype: StepInput
        """
        input_record = self._read_transaction(tx.get_task_input, task=task, input_id=input_id)
        return _reconstruct_task_input(input_record["i"])

    def set_task_input(self, task, input_id, value):
        """Set the value of a task input.

        :param task: the task whose input to set
        :type task: Task
        :param input_id: the ID of the input
        :type input_id: str
        :param value: str or int or float
        """
        self._write_transaction(tx.set_task_input, task=task, input_id=input_id, value=value)

    def get_task_output(self, task, output_id):
        """Get a task output object.

        :param task: the task whose output to retrieve
        :type task: Task
        :param output_id: the ID of the output
        :type output_id: str
        :rtype: StepOutput
        """
        output_record = self._read_transaction(tx.get_task_output, task=task, output_id=output_id)
        return _reconstruct_task_output(output_record["o"])

    def set_task_output(self, task, output_id, value):
        """Set the value of a task output.

        :param task: the task whose output to set
        :type task: Task
        :param output_id: the ID of the output
        :type output_id: str
        :param value: the output value to set
        :type value: str or int or float
        """
        self._write_transaction(tx.set_task_output, task=task, output_id=output_id, value=value)

    def set_task_input_type(self, task, input_id, type_):
        """Set the type of a task input.

        :param task: the task whose input type to set
        :type task: Task
        :param input_id: the ID of the input
        :type input_id: str
        :param type_: the input type to set
        :param type_: str
        """
        self._write_transaction(tx.set_task_input_type, task=task, input_id=input_id, type_=type_)

    def set_task_output_glob(self, task, output_id, glob):
        """Set the glob of a task output.

        :param task: the task whose output to set
        :type task: Task
        :param output_id: the ID of the output
        :type output_id: str
        :param glob: the output glob to set
        :type glob: str
        """
        self._write_transaction(tx.set_task_output_glob, task=task, output_id=output_id, glob=glob)

    def workflow_completed(self):
        """Determine if a workflow in the Neo4j database has completed.

        A workflow has completed if each of its final task nodes have state 'COMPLETED'.
        :rtype: bool
        """
        return self._read_transaction(tx.final_tasks_completed)

    def empty(self):
        """Determine if the Neo4j database is empty.

        :rtype: bool
        """
        return self._read_transaction(tx.is_empty)

    def cleanup(self):
        """Clean up all data in the Neo4j database."""
        self._write_transaction(tx.cleanup)

    def close(self):
        """Close the connection to the Neo4j database."""
        self._driver.close()

    def _get_task_data_tuples(self, task_records):
        """Get a list of (task_record, hints, requirements, inputs, outputs) tuples.

        :param task_records: the database records of the tasks
        :type task_records: BoltStatementResult
        :rtype: list of (BoltStatementResult, list of Hint, list of Requirement)
        """
        with self._driver.session() as session:
            trecords = list(task_records)
            hint_records = [session.read_transaction(tx.get_task_hints,
                            task_id=rec["t"]["id"]) for rec in trecords]
            req_records = [session.read_transaction(tx.get_task_requirements,
                           task_id=rec["t"]["id"]) for rec in trecords]
            input_records = [session.read_transaction(tx.get_task_inputs,
                             task_id=rec["t"]["id"]) for rec in trecords]
            output_records = [session.read_transaction(tx.get_task_outputs,
                              task_id=rec["t"]["id"]) for rec in trecords]

        hints = [_reconstruct_hints(hint_record) for hint_record in hint_records]
        reqs = [_reconstruct_requirements(req_record) for req_record in req_records]
        inputs = [_reconstruct_task_inputs(input_record) for input_record in input_records]
        outputs = [_reconstruct_task_outputs(output_record) for output_record in output_records]

        return list(zip(trecords, hints, reqs, inputs, outputs))

    def _read_transaction(self, tx_fun, **kwargs):
        """Run a Neo4j read transaction.

        :param tx_fun: the transaction function to run
        :type tx_fun: function
        :param kwargs: optional parameters for the transaction function
        """
        # Wrapper for neo4j.Session.read_transaction
        with self._driver.session() as session:
            result = session.read_transaction(tx_fun, **kwargs)
        return result

    def _write_transaction(self, tx_fun, **kwargs):
        """Run a Neo4j write transaction.

        :param tx_fun: the transaction function to run
        :type tx_fun: function
        :param kwargs: optional parameters for the transaction function
        """
        # Wrapper for neo4j.Session.write_transaction
        with self._driver.session() as session:
            session.write_transaction(tx_fun, **kwargs)


def _reconstruct_requirements(req_records):
    """Reconstruct requirements by their records retrieved from Neo4j.

    :param req_records: the database record of the requirements
    :type req_records: BoltStatementResult
    :rtype: list of Requirement
    """
    recs = [req_record["r"] for req_record in req_records]
    return [Requirement(rec["class"], {k: v for k, v in rec.items() if k != "class"})
            for rec in recs]


def _reconstruct_hints(hint_records):
    """Reconstruct hints by their records retrieved from Neo4j.

    :param hint_records: the database record of the hints
    :type hint_records: BoltStatementResult
    :rtype: list of Hint
    """
    recs = [hint_record["h"] for hint_record in hint_records]
    return [Hint(rec["class"], {k: v for k, v in rec.items() if k != "class"}) for rec in recs]


def _reconstruct_workflow_inputs(input_records):
    """Reconstruct workflow inputs by their records retrieved from Neo4j.

    :param input_records: the database record of the inputs
    :type input_records: BoltStatementResult
    :rtype: list of InputParameter
    """
    recs = [input_record["i"] for input_record in input_records]
    return [InputParameter(rec["id"], rec["type"], rec["value"]) for rec in recs]


def _reconstruct_workflow_outputs(output_records):
    """Reconstruct workflow outputs by their records retrieved from Neo4j.

    :param output_records: the database record of the outputs
    :type output_records: BoltStatementResult
    :rtype: list of OutputParameter
    """
    recs = [output_record["o"] for output_record in output_records]
    return [OutputParameter(rec["id"], rec["type"], rec["value"], rec["source"]) for rec in recs]


def _reconstruct_task_inputs(input_records):
    """Reconstruct task inputs by their records retrieved from Neo4j.

    :param input_records: the database record of the inputs
    :type input_records: BoltStatementResult
    :rtype: list of StepInput
    """
    recs = [input_record["i"] for input_record in input_records]
    return [_reconstruct_task_input(rec) for rec in recs]


def _reconstruct_task_input(rec):
    """Reconstruct a task input by its record retrieved from Neo4j.

    :param rec: the database record of the input
    :type rec: BoltStatementResult
    :rtype: StepInput
    """
    return StepInput(rec["id"], rec["type"], rec["value"], rec["default"], rec["source"],
                     rec["prefix"], rec["position"], rec["value_from"])


def _reconstruct_task_outputs(output_records):
    """Reconstruct task outputs by their records retrieved from Neo4j.

    :param output_records: the database record of the outputs
    :type output_records: BoltStatementResult
    :rtype: list of StepOutput
    """
    recs = [output_record["o"] for output_record in output_records]
    return [_reconstruct_task_output(rec) for rec in recs]


def _reconstruct_task_output(rec):
    """Reconstruct a task output by its record retrieved from Neo4j.

    :param rec: the database record of the output
    :type rec: BoltStatementResult
    :rtype: StepOutput
    """
    return StepOutput(rec["id"], rec["type"], rec["value"], rec["glob"])


def _reconstruct_workflow(workflow_record, hints, requirements, inputs, outputs):
    """Reconstruct a Workflow object by its record retrieved from Neo4j.

    :param workflow_record: the database record of the workflow
    :type workflow_record: BoltStatementResult
    :param hints: the workflow hints
    :type hints: list of Hint
    :param requirements: the workflow requirements
    :type requirements: list of Requirement
    :param inputs: the workflow inputs
    :type inputs: list of InputParameter
    :param outputs: the workflow outputs
    :type outputs: list of OutputParameter
    :rtype: Workflow
    """
    rec = workflow_record["w"]
    return Workflow(name=rec["name"], hints=hints, requirements=requirements, inputs=inputs,
                    outputs=outputs, workflow_id=rec["id"])


def _reconstruct_task(task_record, hints, requirements, inputs, outputs):
    """Reconstruct a Task object by its record retrieved from Neo4j.

    :param task_record: the database record of the task
    :type task_record: BoltStatementResult
    :param hints: the task hints
    :type hints: list of Hint
    :param requirements: the task requirements
    :type requirements: list of Requirement
    :param inputs: the task inputs
    :type inputs: list of StepInput
    :param outputs: the task outputs
    :type outputs: list of StepOutput
    :rtype: Task
    """
    rec = task_record["t"]
    return Task(name=rec["name"], base_command=rec["base_command"], hints=hints,
                requirements=requirements, inputs=inputs, outputs=outputs, stdout=rec["stdout"],
                stderr=rec["stderr"], workflow_id=rec["workflow_id"], task_id=rec["id"])


def _reconstruct_metadata(metadata_record):
    """Reconstruct a dict containing the job description metadata retrieved from Neo4j.

    :param metadata_record: the database record of the metadata
    :type metadata_record: BoltStatementResult
    :param keys: the metadata keys to retrieve from the record
    :type keys: iterable of str
    :rtype: dict
    """
    rec = metadata_record["m"]
    return {key: val for key, val in rec.items() if key != "state"}

# pylama:ignore=E1129



"""Configuration validation."""


class ConfigError(Exception):
    """Configuration error class."""


class ConfigValidator:
    """Config validation/schema class for managing configuration."""

    def __init__(self, description):
        """Construct the config validator."""
        self.description = description
        self._section_order = []
        self._sections = {}
        self._options = {}

    def validate(self, conf):
        """Validate a config and return a new config with defaults and values converted."""
        new_conf = {}
        # first validate each option
        for sec_name in conf:
            new_conf[sec_name] = {}
            section = conf[sec_name]
            if sec_name not in self._sections:
                raise ValueError(f'section "{sec_name}" is not defined')
            # check that this section is valid in this context
            dep = self._sections[sec_name].depends_on
            if not self._validate_section(conf, depends_on=dep):
                raise ValueError(
                    f"section '{sec_name}' requires that {dep[0]}::{dep[1]} == '{dep[2]}'"
                )
            # validate each option
            for opt_name in section:
                key = (sec_name, opt_name)
                if key not in self._options:
                    raise ValueError(
                        f"option '{opt_name}' in section '{sec_name}' is not defined"
                    )
                option = self._options[key]
                try:
                    new_conf[sec_name][opt_name] = option.validate(section[opt_name])
                except ValueError as err:
                    raise ValueError(
                        f'{sec_name}::{opt_name}: {err.args[0]}'
                    ) from None
        # then check required options
        for key in self._options:
            if not self._validate_section(conf, self._sections[key[0]].depends_on):
                # skip invalid sections
                continue
            if key[0] not in conf or key[1] not in conf[key[0]]:
                raise ValueError(
                    f"option '{key[1]}' in section '{key[0]}' is required but has not been set"
                )

        return new_conf

    def _validate_section(self, conf, depends_on):
        """Ensure that this section is valid in this context (check depends relations)."""
        if depends_on is None:
            return True
        if depends_on[0] not in conf or depends_on[1] not in conf[depends_on[0]]:
            return False
        # value must match for the section to be valid
        return conf[depends_on[0]][depends_on[1]] == depends_on[2]

    def is_section_valid(self, cur_conf, sec_name):
        """Determine if given the current configuration, sec_name is valid.

        sec_name is assumed to be the name of a valid section.
        """
        depends_on = self._sections[sec_name].depends_on
        return self._validate_section(cur_conf, depends_on)

    def section(self, sec_name, info, depends_on=None):
        """Define a configuration section."""
        if sec_name in self._sections:
            raise ConfigError(f"section '{sec_name}' has already been defined")
        self._sections[sec_name] = ConfigSection(info, depends_on=depends_on)
        # save insertion order
        self._section_order.append(sec_name)

    def option(self, sec_name, opt_name, *args, **kwargs):
        """Define a configuration option."""
        if sec_name not in self._sections:
            raise ConfigError(f"section '{sec_name}' has not been defined")
        key = (sec_name, opt_name)
        if key in self._options:
            raise ConfigError(
                f"option '{opt_name}' in section '{sec_name}' has already been defined"
            )
        try:
            self._options[key] = ConfigOption(*args, **kwargs)
        except ConfigError as err:
            raise ConfigError(f'{sec_name}::{opt_name}: {err.args[0]}') from None

    @property
    def sections(self):
        """Return all sections in order as list of tuples (sec_name, section)."""
        sec_names = list(self._section_order)
        s = set()
        order = []
        # Ensure dependency ordering
        for sec_name in sec_names:
            if sec_name in s:
                continue
            section = self._sections[sec_name]
            depends_on = section.depends_on
            if depends_on is not None and depends_on[0] not in s:
                s.add(depends_on[0])
                order.append((depends_on[0], self._sections[depends_on[0]]))
            s.add(sec_name)
            order.append((sec_name, section))
        return order

    def options(self, sec_name):
        """Return all options for a given section."""
        return [(key[1], value) for key, value in self._options.items() if key[0] == sec_name]


class ConfigSection:
    """Config section."""

    def __init__(self, info, depends_on=None):
        """Config section constructor."""
        self.info = info
        # (sec_name, key, value)
        self.depends_on = depends_on


class ConfigOption:
    """Config option/validation class."""

    def __init__(self, info, validator=str, choices=None, attrs=None):
        """Construct the config option class."""
        self.info = info
        self.choices = choices
        # attrs is designed to hold any extra attribute that could be useful
        self.attrs = attrs
        self._validator = validator

    def validate(self, value):
        """Validate the value and return it."""
        if self.choices is not None:
            if value not in self.choices:
                raise ValueError(f"value must be one of {self.choices}; found '{value}'")
            return value
        return self._validator(value)



"""Scheduler REST test cases.

Tests of the REST interface for BEE Scheduler.
"""
import tempfile
import pytest

from beeflow.scheduler.scheduler import create_app


SCHEDULER_TEST_PORT = '5100'


def endpoint(*pargs):
    """Return the endpoint url for a resource."""
    return '/'.join(['bee_sched/v1', *pargs])


@pytest.fixture(scope='function')
def scheduler(mocker):
    """Fixture code to setup a new scheduler per test function.

    Start a new scheduler as a subprocess for each test function.
    """
    app = create_app()
    app.config.update({
        'TESTING': True,
    })
    with tempfile.NamedTemporaryFile() as fp:
        mocker.patch('beeflow.scheduler.scheduler.get_db_path', return_value=fp.name)
        yield app.test_client()


def test_schedule_job_no_resources(scheduler):
    """Test scheduling a job with no resources.

    Test scheduling a job with no resources.
    :param scheduler: connection returned by fixture
    :type scheduler: Connection
    """
    conn = scheduler
    workflow_name = 'test-workflow'
    task1 = {
        'workflow_name': workflow_name,
        'task_name': 'test-task',
        'requirements': {
            'max_runtime': 1,
        },
    }
    req = conn.put(endpoint('workflows', workflow_name, 'jobs'), json=[task1])

    assert req.status_code == 200
    data = req.json
    assert len(data) == 1
    assert data[0]['workflow_name'] == workflow_name
    assert data[0]['task_name'] == 'test-task'
    assert data[0]['requirements']['max_runtime'] == 1
    assert data[0]['allocations'] == []


def test_schedule_job_one_resource(scheduler):
    """Test scheduling a job with one resource.

    Test scheduling a job with one resource.
    :param scheduler: connection returned by fixture
    :type scheduler: Connection
    """
    conn = scheduler
    # with scheduler() as url:
    # Create a single resource
    resource1 = {
        'id_': 'resource-1',
        'nodes': 10,
    }
    req = conn.put(endpoint('resources'), json=[resource1])

    assert req.status_code == 200
    assert req.json == 'created 1 resource(s)'

    workflow_name = 'test-workflow'
    task1 = {
        'workflow_name': 'test-workflow',
        'task_name': 'test-task',
        'requirements': {
            'max_runtime': 1,
        },
    }
    req = conn.put(endpoint('workflows', workflow_name, 'jobs'), json=[task1])

    assert req.status_code == 200
    data = req.json
    assert len(data) == 1
    assert data[0]['workflow_name'] == 'test-workflow'
    assert data[0]['task_name'] == 'test-task'
    assert data[0]['requirements']['max_runtime'] == 1
    assert len(data[0]['allocations']) == 1
    assert data[0]['allocations'][0]['id_'] == 'resource-1'
    assert data[0]['allocations'][0]['nodes'] == 1
    assert data[0]['allocations'][0]['start_time'] == 0
    assert data[0]['allocations'][0]['max_runtime'] == 1


def test_schedule_job_two_resources(scheduler):
    """Test scheduling a job with two resources.

    Test scheduling a job with two resources.
    :param scheduler: connection returned by fixture
    :type scheduler: Connection
    """
    conn = scheduler
    # with scheduler() as url:
    # Create a single resource
    resource1 = {
        'id_': 'resource-1',
        'nodes': 10,
    }
    resource2 = {
        'id_': 'resource-2',
        'nodes': 64,
    }
    req = conn.put(endpoint('resources'), json=[resource1, resource2])

    assert req.status_code == 200
    assert req.json == 'created 2 resource(s)'

    workflow_name = 'test-workflow'
    task1 = {
        'workflow_name': 'test-workflow',
        'task_name': 'test-task',
        'requirements': {
            'max_runtime': 1,
        },
    }
    req = conn.put(endpoint('workflows', workflow_name, 'jobs'), json=[task1])

    assert req.status_code == 200
    data = req.json
    assert len(data) == 1
    assert data[0]['workflow_name'] == 'test-workflow'
    assert data[0]['task_name'] == 'test-task'
    assert data[0]['requirements']['max_runtime'] == 1
    assert len(data[0]['allocations']) > 0


def test_schedule_multi_job_two_resources(scheduler):
    """Test scheduling multiple jobs with two resources.

    Test scheduling multiple jobs with two resources.
    :param scheduler: connection returned by fixture
    :type scheduler: Connection
    """
    conn = scheduler
    # with scheduler() as url:
    # Create a single resource
    resource1 = {
        'id_': 'resource-0',
        'nodes': 10,
    }
    resource2 = {
        'id_': 'resource-1',
        'nodes': 16,
    }
    req = conn.put(endpoint('resources'), json=[resource1, resource2])

    assert req.status_code == 200
    assert req.json == 'created 2 resource(s)'

    workflow_name = 'test-workflow'
    task1 = {
        'workflow_name': 'test-workflow',
        'task_name': 'test-task-0',
        'requirements': {
            'max_runtime': 1,
        },
    }
    task2 = {
        'workflow_name': 'test-workflow',
        'task_name': 'test-task-1',
        'requirements': {
            'max_runtime': 1,
        },
    }
    task3 = {
        'workflow_name': 'test-workflow',
        'task_name': 'test-task-2',
        'requirements': {
            'max_runtime': 4,
            'nodes': 16,
        },
    }
    req = conn.put(endpoint('workflows', workflow_name, 'jobs'),
                   json=[task1, task2, task3])

    assert req.status_code == 200
    data = req.json
    assert len(data) == 3
    assert data[0]['workflow_name'] == 'test-workflow'
    assert data[0]['task_name'] == 'test-task-0'
    assert data[0]['requirements']['max_runtime'] == 1
    assert len(data[0]['allocations']) > 0
    # Ensure proper scheduled time
    assert data[0]['allocations'][0]['start_time'] < 6
    assert data[1]['workflow_name'] == 'test-workflow'
    assert data[1]['task_name'] == 'test-task-1'
    assert data[1]['requirements']['max_runtime'] == 1
    assert len(data[1]['allocations']) > 0
    # Ensure proper scheduled time
    assert data[1]['allocations'][0]['start_time'] < 6
    assert data[2]['workflow_name'] == 'test-workflow'
    assert data[2]['task_name'] == 'test-task-2'
    assert data[2]['requirements']['max_runtime'] == 4
    assert data[2]['requirements']['nodes'] == 16
    assert len(data[2]['allocations']) > 0
    # Ensure proper scheduled time
    assert data[2]['allocations'][0]['start_time'] < 6


# Ignore R1732: This suggestion about using `with` doesn't apply here.
# Ignore W0621: These are fixtures; it's supposed to work this way.
# pylama:ignore=R1732,W0621



"""Utility functions for wf_manager resources."""

import os
import shutil
import socket
import requests
import jsonpickle

from beeflow.common import log as bee_logging
from beeflow.wf_manager.common import wf_db
from beeflow.common.config_driver import BeeConfig as bc
from beeflow.common.gdb_interface import GraphDatabaseInterface
from beeflow.common.gdb.neo4j_driver import Neo4jDriver
from beeflow.common.wf_interface import WorkflowInterface
from beeflow.common.connection import Connection


log = bee_logging.setup(__name__)


def get_bee_workdir():
    """Get the bee workflow directory from the configuration file."""
    return os.path.expanduser('~/.beeflow')


def get_workflows_dir():
    """Get the workflows script directory from beeflow."""
    bee_workdir = get_bee_workdir()
    workflows_dir = os.path.join(bee_workdir, 'workflows')
    return workflows_dir


def get_workflow_dir(wf_id):
    """Get the workflow script dir for a particular workflow."""
    return os.path.join(get_workflows_dir(), wf_id)


def create_workflow_dir(wf_id):
    """Create the workflows directory."""
    os.makedirs(get_workflow_dir(wf_id))


def create_current_run_dir():
    """Create directory to store current run GDB bind info."""
    bee_workdir = get_bee_workdir()
    current_run_dir = os.path.join(bee_workdir, 'current_run')
    os.makedirs(current_run_dir)


def remove_current_run_dir():
    """Remove current run directory."""
    bee_workdir = get_bee_workdir()
    current_run_dir = os.path.join(bee_workdir, 'current_run')
    if os.path.exists(current_run_dir):
        shutil.rmtree(current_run_dir)


def remove_wf_dir(wf_id):
    """Remove a workflow directory."""
    bee_workdir = get_bee_workdir()
    workflows_dir = os.path.join(bee_workdir, 'workflows', wf_id)
    if os.path.exists(workflows_dir):
        shutil.rmtree(workflows_dir)
    # wf_db.delete_workflow(wf_id)


def create_wf_metadata(wf_id, wf_name):
    """Create workflow metadata files."""
    create_wf_name(wf_id, wf_name)
    create_wf_status(wf_id)
    # wf_db.add_workflow(wf_id, wf_name, 'Pending')


def create_wf_name(wf_id, wf_name):
    """Create workflow name metadata file."""
    bee_workdir = get_bee_workdir()
    workflows_dir = os.path.join(bee_workdir, 'workflows', wf_id)
    name_path = os.path.join(workflows_dir, 'bee_wf_name')
    with open(name_path, 'w', encoding="utf8") as name:
        name.write(wf_name)


def create_wf_status(wf_id):
    """Create workflow status metadata file."""
    update_wf_status(wf_id, 'Pending')


def update_wf_status(wf_id, status_msg):
    """Update workflow status metadata file."""
    bee_workdir = get_bee_workdir()
    workflows_dir = os.path.join(bee_workdir, 'workflows', wf_id)
    status_path = os.path.join(workflows_dir, 'bee_wf_status')
    with open(status_path, 'w', encoding="utf8") as status:
        status.write(status_msg)
    wf_db.update_workflow_state(wf_id, status_msg)


def read_wf_status(wf_id):
    """Read workflow status metadata file."""
    bee_workdir = get_bee_workdir()
    workflows_dir = os.path.join(bee_workdir, 'workflows', wf_id)
    status_path = os.path.join(workflows_dir, 'bee_wf_status')
    with open(status_path, 'r', encoding="utf8") as status:
        wf_status = status.readline()
    return wf_status


def create_wf_namefile(wf_name, wf_id):
    """Create workflow name metadata file."""
    bee_workdir = get_bee_workdir()
    workflows_dir = os.path.join(bee_workdir, 'workflows', wf_id)
    name_path = os.path.join(workflows_dir, 'bee_wf_name')
    with open(name_path, 'w', encoding="utf8") as name:
        name.write(wf_name)


def get_workflow_interface(wf_id):
    """Instantiate and return workflow interface object."""
    bolt_port = wf_db.get_bolt_port(wf_id)
    try:
        driver = Neo4jDriver(user="neo4j", bolt_port=bolt_port,
                             db_hostname=bc.get("graphdb", "hostname"),
                             password=bc.get("graphdb", "dbpass"))
        iface = GraphDatabaseInterface(driver)
        wfi = WorkflowInterface(iface)
    except KeyError:
        log.error('The default way to load WFI didnt work')
        # wfi = WorkflowInterface()
    return wfi


def tm_url():
    """Get Task Manager url."""
    # tm_listen_port = bc.get('task_manager', 'listen_port')
    tm_listen_port = wf_db.get_tm_port()
    task_manager = "bee_tm/v1/task/"
    return f'http://127.0.0.1:{tm_listen_port}/{task_manager}'


# Base URLs for the TM and the Scheduler
TM_URL = "bee_tm/v1/task/"
SCHED_URL = "bee_sched/v1/"


def _connect_tm():
    """Return a connection to the TM."""
    return Connection(bc.get('task_manager', 'socket'))


def sched_url():
    """Get Scheduler url."""
    scheduler = "bee_sched/v1/"
    # sched_listen_port = bc.get('scheduler', 'listen_port')
    sched_listen_port = wf_db.get_sched_port()
    return f'http://127.0.0.1:{sched_listen_port}/{scheduler}'


def _connect_scheduler():
    """Return a connection to the Scheduler."""
    return Connection(bc.get('scheduler', 'socket'))


def _resource(component, tag=""):
    """Access Task Manager or Scheduler."""
    if component == "tm":
        url = TM_URL + str(tag)
    elif component == "sched":
        url = SCHED_URL + str(tag)
    return url


# Submit tasks to the TM
# pylama:ignore=W0613
def submit_tasks_tm(wf_id, tasks, allocation):
    """Submit a task to the task manager."""
    wfi = get_workflow_interface(wf_id)
    for task in tasks:
        metadata = wfi.get_task_metadata(task)
        task.workdir = metadata['workdir']
    # Serialize task with json
    tasks_json = jsonpickle.encode(tasks)
    # Send task_msg to task manager
    names = [task.name for task in tasks]
    log.info(f"Submitted {names} to Task Manager")
    try:
        conn = _connect_tm()
        resp = conn.post(_resource('tm', "submit/"), json={'tasks': tasks_json},
                         timeout=5)
    except requests.exceptions.ConnectionError:
        log.error('Unable to connect to task manager to submit tasks.')
        return

    if resp.status_code != 200:
        log.info(f"Submit task to TM returned bad status: {resp.status_code}")


def tasks_to_sched(tasks):
    """Convert gdb tasks to sched tasks."""
    sched_tasks = []
    for task in tasks:
        sched_task = {
            'workflow_name': 'workflow',
            'task_name': task.name,
            'requirements': {
                'max_runtime': 1,
                'nodes': 1
            }
        }
        sched_tasks.append(sched_task)
    return sched_tasks


def submit_tasks_scheduler(tasks):
    """Submit a list of tasks to the scheduler."""
    sched_tasks = tasks_to_sched(tasks)
    # The workflow name will eventually be added to the wfi workflow object
    try:
        conn = _connect_scheduler()
        resp = conn.put(_resource('sched', "workflows/workflow/jobs"), json=sched_tasks,
                        timeout=5)
    except requests.exceptions.ConnectionError:
        log.error('Unable to connect to scheduler to submit tasks.')
        return "Did not work"

    if resp.status_code != 200:
        log.info(f"Something bad happened {resp.status_code}")
        return "Did not work"
    return resp.json()


def get_open_port():
    """Return an open ephemeral port."""
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    s.listen(1)
    port = s.getsockname()[1]
    s.close()
    return port


def schedule_submit_tasks(wf_id, tasks):
    """Schedule and then submit tasks to the TM."""
    # Submit ready tasks to the scheduler
    allocation = submit_tasks_scheduler(tasks)  #NOQA
    # Submit tasks to TM
    submit_tasks_tm(wf_id, tasks, allocation)
